/*-------------------------------------------------------------------------+
| Copyright 2016 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.safetyfunctionarch;

import static org.fortiss.af3.exploration.alg.util.AF3Utils.getChannelNumbersSafetyArch;
import static org.fortiss.af3.exploration.alg.util.AF3Utils.getTriggerNumbersSafetyArch;
import static org.fortiss.af3.exploration.alg.util.AF3Utils.isDiagConnectedSafetyArch;
import static org.fortiss.af3.exploration.alg.util.AF3Utils.isDiagPresentSafetyArch;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.PhenotypeBase;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.operator.copy.Copyable;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ISafetyFunctionAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.random.RandomExploration;
import org.fortiss.af3.exploration.alg.service.module.IExplorationEncoding;
import org.fortiss.af3.safety.model.SafetyArchitecture;
import org.opt4j.core.Genotype;

/**
 * Encodes the design of safety functions that are contained in a logical architecture. It consists
 * of the {@link SafetyFunctionArchEntry}s of all identified safety function of the input model.
 * 
 * @author diewald
 */
public class SafetyFunctionArchEncoding extends PhenotypeBase
		implements IExplorationEncoding, Genotype, Copyable<SafetyFunctionArchEncoding> {

	/** Random generator */
	private RandomExploration rand = RandomExploration.getInstance();

	/**
	 * Collection that consists of safety function design entries for each safety function that is
	 * present in the input models.
	 */
	private List<SafetyFunctionArchEntry> safetyFunctionArch = new ArrayList<>();

	/**
	 * Constructor for creating a {@link SafetyFunctionArchEncoding} with entries for each safety
	 * function defined in the input models. The selected entries can be either {@code randomize}d
	 * or default to a 1oo1 design.
	 */
	public SafetyFunctionArchEncoding(Collection<ISafetyFunctionAdapter<?>> safetyFunctions,
			boolean initialize, boolean randomize) {
		if(initialize) {
			for(ISafetyFunctionAdapter<?> currSF : safetyFunctions) {
				if(randomize) {
					safetyFunctionArch.add(createRandomSFDesign(currSF));
				} else {
					safetyFunctionArch.add(createDefaultSFDesign(currSF));
				}
			}
		}
	}

	/**
	 * Constructor for creating a {@link SafetyFunctionArchEncoding} with the given encoding. Each
	 * {@link SafetyFunctionArchEntry} will be cloned.
	 */
	public SafetyFunctionArchEncoding(Collection<SafetyFunctionArchEntry> safetyFunctionDesigns) {
		for(SafetyFunctionArchEntry entry : safetyFunctionDesigns) {
			this.safetyFunctionArch.add(entry.clone());
		}
	}

	/** Creates a 1oo1 {@link SafetyFunctionArchEntry}. */
	public SafetyFunctionArchEntry createDefaultSFDesign(ISafetyFunctionAdapter<?> safetyFunction) {
		return new SafetyFunctionArchEntry(safetyFunction, 1, 0, false, false);
	}

	/**
	 * Creates a {@link SafetyFunctionArchEntry} with randomized parameters subject to the
	 * limitations defined in the given {@link ISafetyFunctionAdapter}.
	 */
	public SafetyFunctionArchEntry createRandomSFDesign(ISafetyFunctionAdapter<?> safetyFunction) {
		List<SafetyArchitecture> sfArch =
				new ArrayList<>(safetyFunction.getAllowedSafetyArchitectures());
		int randSFArchIdx = rand.nextInt(sfArch.size());
		SafetyArchitecture selectedSFArch = sfArch.get(randSFArchIdx);

		int selChannelNum = getChannelNumbersSafetyArch(selectedSFArch);
		int triggerNum = getTriggerNumbersSafetyArch(selectedSFArch);
		boolean diagPresent = isDiagPresentSafetyArch(selectedSFArch).orElse(false);
		boolean diagUnitConnected = isDiagConnectedSafetyArch(selectedSFArch).orElse(false);

		return new SafetyFunctionArchEntry(safetyFunction, selChannelNum, triggerNum, diagPresent,
				diagUnitConnected);
	}

	/**
	 * Adds a {@link SafetyFunctionArchEntry} to the encoding of all safety functions in the
	 * system.
	 */
	public boolean addEntry(SafetyFunctionArchEntry entry) {
		return safetyFunctionArch.add(entry);
	}

	/**
	 * Returns the {@link SafetyFunctionArchEntry} that is located at the given {@code index} of the
	 * encoding.
	 */
	public SafetyFunctionArchEntry getEntry(int index) {
		if(index < 0 || index >= safetyFunctionArch.size()) {
			return null;
		}
		return safetyFunctionArch.get(index);
	}

	/**
	 * Returns the {@link SafetyFunctionArchEntry}s that compose {@code this}
	 * {@link SafetyFunctionArchEncoding}.
	 */
	public Collection<SafetyFunctionArchEntry> getAllEntries() {
		return safetyFunctionArch;
	}

	/**
	 * Returns the {@link SafetyFunctionArchEntry} that represents the chosen design of the given
	 * {@link ISafetyFunctionAdapter}.
	 */
	public Optional<SafetyFunctionArchEntry> getDesignOf(ISafetyFunctionAdapter<?> safetyFunction) {
		return safetyFunctionArch.stream()
				.filter(sfd -> sfd.getSafetyFunctionAdapter() == safetyFunction).findAny();
	}

	/** {@inheritDoc} */
	@Override
	public int size() {
		return safetyFunctionArch.size();
	}

	/** {@inheritDoc} */
	@Override
	public <G extends Genotype> G newInstance() {
		Collection<ISafetyFunctionAdapter<?>> sfAdapters = safetyFunctionArch.stream()
				.map(sf -> sf.getSafetyFunctionAdapter()).collect(Collectors.toList());
		@SuppressWarnings("unchecked") G sfArchEncoding =
				(G)new SafetyFunctionArchEncoding(sfAdapters, true, true);
		return sfArchEncoding;
	}

	/** {@inheritDoc} */
	@Override
	public SafetyFunctionArchEncoding copy() {
		return new SafetyFunctionArchEncoding(safetyFunctionArch);
	}

	/** {@inheritDoc} */
	@Override
	public String toString() {
		return getAllEntries().toString();
	}
}
