/*-------------------------------------------------------------------------+
| Copyright 2018 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.evaluate;

import static org.fortiss.af3.exploration.alg.util.ExplorationAlgUtils.getAnnotatedMethod;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.evaluate.annotations.Evaluate;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.evaluate.annotations.Evaluates;
import org.fortiss.af3.exploration.alg.dse.module.IEvaluatorModule;
import org.fortiss.af3.exploration.alg.exception.ExplorationException;
import org.fortiss.af3.exploration.model.ExplorationTarget;
import org.opt4j.core.DoubleValue;
import org.opt4j.core.Objective;
import org.opt4j.core.Value;

import com.google.inject.AbstractModule;
import com.google.inject.Injector;
import com.google.inject.Key;

/**
 * Base class for {@link ICompositeEvaluator}s that may contain additional sub-evaluators.
 * 
 * @author diewald
 */
public abstract class CompositeEvaluator<R, P> extends AbstractModule
		implements ICompositeEvaluator<R> {

	/** Corresponding Opt4J {@link Objective}of an {@link ExplorationTarget}. */
	protected Objective objective;

	/**
	 * References all {@link IEvaluatorModule}s composing {@code this} {@link CompositeEvaluator}.
	 */
	protected Set<IEvaluatorModule<?, ?, ?>> subEvaluators = Collections.emptySet();

	/**
	 * References the {@link Injector} that holds bindings for the objects used by all Evaluator
	 * instances and threads.
	 */
	private Injector centralEvaluatorInjector;

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public Value<R> evaluate() {
		// First, we create another injector such that the Provider for the evaluation result
		// type is encapsulated from the other instances. Otherwise, we would have multiple
		// Provides methods for the same type, which guice does not permit.
		EvaluatorInjectorProvider evalInjectorProvider =
				centralEvaluatorInjector.getInstance(EvaluatorInjectorProvider.class);
		Injector localEvaluationInjector =
				evalInjectorProvider.getInjector().createChildInjector(this);

		List<Value<R>> subResults = new ArrayList<>();
		for(IEvaluatorModule<R, ?, ?> subEvaluator : getSubEvaluators()) {
			subEvaluator.setEvaluationInjectorIfAbsent(centralEvaluatorInjector);
			subResults.add(subEvaluator.evaluate());
		}

		// Safe due to type erasure. The max double value indicates a wrong evaluation.
		Value<R> evalResult = (Value<R>)new DoubleValue(Double.MAX_VALUE);
		try {
			Method evalMethod = getAnnotatedMethod(getClass(), Evaluate.class);
			Class<? extends Value<?>> evalType =
					(Class<? extends Value<?>>)evalMethod.getReturnType();
			evalResult = (Value<R>)localEvaluationInjector.getInstance(
					Key.get(evalType, Evaluates.eval((Class<ICompositeEvaluator<?>>)getClass())));
		} catch(Exception e) {
			e.printStackTrace();
			// The Opt4J framework only allows us to throw runtime exceptions.
			throw new RuntimeException(new ExplorationException("The evaluatior " +
					this.toString() + " produced an error while performing the evaluation.", e));
		}

		return evalResult;
	}

	/** {@inheritDoc} */
	@Override
	public Objective getObjective() {
		return objective;
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public Set<IEvaluatorModule<R, ?, ?>> getAllSubEvaluators() {
		Set<IEvaluatorModule<R, ?, ?>> allSubEvaluators = new HashSet<>();
		// Get the direct sub-Evaluators.
		allSubEvaluators.addAll(getSubEvaluators());

		// Recursively fetch the nested sub-Evaluators.
		for(ICompositeEvaluator<?> curCompositeEval : getSubEvaluators()) {
			allSubEvaluators.addAll((Set<? extends IEvaluatorModule<R, ?, ?>>)curCompositeEval
					.getAllSubEvaluators());
		}

		return allSubEvaluators;
	}

	/** {@inheritDoc} */
	@Override
	public synchronized void setEvaluationInjectorIfAbsent(Injector evaluationInjector) {
		if(this.centralEvaluatorInjector == null) {
			this.centralEvaluatorInjector = evaluationInjector;
		}
	}

	/** Returns the set of sub-evaluators of the current one. */
	protected abstract Set<IEvaluatorModule<R, ?, ?>> getSubEvaluators();
}
