/*-------------------------------------------------------------------------+
| Copyright 2016 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.modeltransformation;

import java.util.Collection;
import java.util.List;
import java.util.Map;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.af3.exploration.alg.exception.TransformationModuleException;
import org.fortiss.af3.exploration.alg.graph.IDependencyModule;
import org.fortiss.af3.exploration.alg.guava.SubClassToInstanceMap;
import org.fortiss.af3.exploration.alg.guava.TypedBiMap;
import org.fortiss.af3.exploration.alg.service.IExplorationExtension;
import org.fortiss.af3.exploration.alg.service.IExplorationTransformationService.TransformationContext;
import org.fortiss.af3.exploration.alg.service.module.ExplorationTransformationInputs;
import org.fortiss.af3.exploration.alg.service.module.IExplorationEncoding;
import org.fortiss.tooling.kernel.service.base.IEObjectAware;

import com.google.common.collect.Multimap;

/**
 * Interface for any transformation of {@link IExplorationEncoding}s of the DSE to EMF models.
 * Each implementing class must declare its output model type via {@link #getOutputType()} and
 * its required input types via {@link #getRequiredTypes()}. {@link ITransformationModule}s are
 * composable such that each may have submodules that can be queried via {@link #getSubModules()}.
 * The submodules must be registered via {@link #registerSubModule(ITransformationModule)}. Further,
 * each {@link ITransformationModule} may declare itself to be optional such that it may be ignored
 * when being launched as a submodule if the required inputs are not available.
 * 
 * The method
 * {@link #transform(TransformationState, TransformationContext, ExplorationTransformationInputs, SubClassToInstanceMap, EObject)}
 * actually triggers the
 * transformation using the
 * module itself and its submodules.
 * 
 * @author diewald
 */
public interface ITransformationModule<T extends EObject>
		extends IEObjectAware<T>, IExplorationExtension, IDependencyModule<T, T> {

	/** Defines the state of a Generation/Transformation module for single element types. */
	enum TransformationState {
			/** Initial state. */
			INITIALIZING("initializing", false, false, false, false, false, true),
			/** Initialization done, e.g., copying the base model. */
			INITIALIZED("initialized", true, false, false, false, false, false),
			/** Module currently replicates elements. */
			REPLICATING("replicating", true, false, false, false, false, true),
			/** Module replicated the defined element types. */
			REPLICATED("replicated", true, true, false, false, false, false),
			/** Module currently instantiates elements. */
			INSTANCIATING("instantiating", true, true, false, false, false, true),
			/** Module instantiated the defined element types. */
			INSTANTIATED("instantiated", true, true, true, false, false, false),
			/** Module currently inserts elements. */
			INSERTING("inserting", true, true, true, false, false, true),
			/** Required elements are inserted. */
			INSERTED("inserted", true, true, true, true, false, false),
			/** Final steps to complete the model. */
			FINALIZING("finalizing", true, true, true, true, false, true),
			/** Model transformation has finished. */
			FINALIZED("finalized", true, true, true, true, true, false);

		/** State name. */
		private final String name;

		/** Initialization status. */
		private final boolean initialized;

		/** Instantiation status. */
		private final boolean instantiated;

		/** Replication status. */
		private final boolean replicated;

		/** Insertion status. */
		private final boolean inserted;

		/** Finalization status. */
		private final boolean finalized;

		/** Whether the Module is currently working. */
		private final boolean processing;

		/** Constructor. */
		TransformationState(String name, boolean initialized, boolean replicated,
				boolean instantiated, boolean inserted, boolean finalized, boolean processing) {
			this.name = name;
			this.initialized = initialized;
			this.replicated = replicated;
			this.instantiated = instantiated;
			this.inserted = inserted;
			this.finalized = finalized;
			this.processing = processing;
		}

		/** Returns whether the module finished the initialization step. */
		public boolean isInitialized() {
			return initialized;
		}

		/** Returns whether the module finished the instantiation step. */
		public boolean isInstantiated() {
			return instantiated;
		}

		/** Returns whether the module finished the replication step. */
		public boolean isReplicated() {
			return replicated;
		}

		/** Returns whether the module finished the insertion step. */
		public boolean isInserted() {
			return inserted;
		}

		/** Returns whether the module transformation is finished. */
		public boolean isFinalized() {
			return finalized;
		}

		/** Returns whether the module is currently performing some operation. */
		public boolean isProcessing() {
			return processing;
		}

		/** {@inheritDoc} */
		@Override
		public String toString() {
			return name;
		}
	}

	/**
	 * Returns the collection of sub-{@link ITransformationModule}s that are registered with
	 * {@code this} {@link ITransformationModule}.
	 */
	List<ITransformationModule<?>> getSubModules();

	/**
	 * Registers the given {@code subModule} with {@code this} {@link ITransformationModule}.
	 */
	void registerSubModule(ITransformationModule<?> subModule);

	/**
	 * Creates another instance of {@code this} {@link ITransformationModule} which will actually
	 * perform the transformation.
	 */
	ITransformationModule<T> createModuleInstance() throws TransformationModuleException;

	/**
	 * Marks the given {@link ITransformationModule} type to not be executed.
	 * 
	 * @param subModule
	 *            The module type to be blacklisted.
	 */
	void addBlacklistedModule(Class<? extends ITransformationModule<?>> subModule);

	/**
	 * Clears the collection of blacklisted modules from {@code this} {@link ITransformationModule}.
	 */
	void resetBlacklistedModules();

	/** Predicates whether the execution of is required or optional. */
	boolean isOptional();

	/**
	 * Updates the given associative maps that allow to trace objects between the original and the
	 * transformed model, and between the entries of and exploration encoding and the elements of
	 * the transformed model. For the update, the associations generated by {@code this}
	 * {@link ITransformationModule} are utilized.
	 * 
	 * @param subModule
	 *            contains the updated association maps.
	 */
	void applyAssociationUpdates(ITransformationModule<?> subModule);

	/**
	 * Returns the model element of a generated/transformed model which is associated with the given
	 * entry of an encoding of the solution found by the DSE.
	 * 
	 * @param encEntry
	 *            The entry of an {@link IExplorationEncoding} to search for.
	 * @param type
	 *            Type of the associated element.
	 * @return The associated {@link EObject} of the generated/transformed model.
	 */
	EObject getModelElementByEncEntry(Object encEntry, Class<EObject> type);

	/**
	 * Returns the model element from the original model which corresponds to the given model
	 * element of the generated/transformed model.
	 * 
	 * @param genElem
	 *            {@link EObject} contained in the generated/transformed model.
	 * @return The corresponding {@link EObject} from the original (input) architecture.
	 */
	EObject getOriginalByGenElement(EObject genElem);

	/**
	 * Removes the given element and its children from its containing model and updates the
	 * association maps accordingly to reflect this change.
	 * 
	 * @param element
	 *            The model element to be removed.
	 */
	void removeGenModelElement(EObject element);

	/**
	 * Returns the transformation stage at which {@code this} {@link ITransformationModule} shall be
	 * launched.
	 */
	TransformationState getActivationStage();

	/**
	 * Returns the collection of parent {@link ITransformationModule} which may launch {@code this}
	 * {@link ITransformationModule} as one of its sub modules.
	 */
	Collection<Class<? extends ITransformationModule<?>>> getParentModules();

	/**
	 * Generates a model of type {@code T} from the input {@code baseModels} and
	 * {@code exploraionEncodings}.
	 * 
	 * @param state
	 *            The state of the transformation engine, e.g., replication.
	 * @param context
	 *            The model context.
	 * @param inputs
	 *            Input types.
	 * @param transformedModels
	 *            Container map of all the transformed models.
	 * @param outputModel
	 *            The output EMF model which includes the transformation result. It is passed to
	 *            each submodule. May be {@code null} if a new model is created in the
	 *            transformation method.
	 * @return Transformed model of type {@code T}.
	 * @throws TransformationModuleException
	 *             If the transformation fails at {@code this} {@link ITransformationModule}.
	 */
	<N extends EObject, S extends EObject> T transform(TransformationState state,
			TransformationContext context, ExplorationTransformationInputs inputs,
			SubClassToInstanceMap<N> transformedModels, EObject outputModel)
			throws TransformationModuleException;

	/**
	 * Returns the associations between elements comprising an exploration encoding and their
	 * corresponding model elements.
	 */
	TypedBiMap<Object, EObject> getEncEntryModelElemAssoc();

	/**
	 * Returns the associative map which relates elements of a generated model to those of the
	 * original input
	 * model.
	 */
	Map<EObject, EObject> getGenToOrigAssocMap();

	/**
	 * Returns the associative map which relates the elements of the original input model to all
	 * elements which from instances of the particular elements.
	 */
	Multimap<EObject, EObject> getOrigToGenAssocMap();

	/**
	 * Returns the associative map which relates the input elements of the current stage to the
	 * output elements of the current stage. Here, the current stage does not differentiate between
	 * the processing and the finish stage.
	 */
	Multimap<EObject, EObject> getGenToGenAssocMap();
}
