/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.modeltransformation.deployment;

import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.af3.component.model.Component;
import org.fortiss.af3.component.model.ComponentArchitecture;
import org.fortiss.af3.deployment.model.Deployment;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.InstantiatedTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.modeltransformation.ITransformationModule;
import org.fortiss.af3.exploration.alg.dse.modeltransformation.base.TransformationStrategyBase;
import org.fortiss.af3.exploration.alg.dse.modeltransformation.comparch.ComponentArchitectureTransformer;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.mapping.IMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.sysmodel.mapping.IMappingEntry;
import org.fortiss.af3.exploration.alg.exception.TransformationModuleException;
import org.fortiss.af3.exploration.alg.guava.SubClassToInstanceMap;
import org.fortiss.af3.exploration.alg.service.IExplorationTransformationService.TransformationContext;
import org.fortiss.af3.exploration.alg.service.module.ExplorationTransformationInputs;
import org.fortiss.af3.platform.model.PlatformArchitecture;

/**
 * Responsible for generating a {@link Deployment} that is compatible with the Deployment-plugin.
 * Thus, a Component -> Platform mapping resulting form the DSE can viewed and edited within the
 * {@link Deployment} editor.
 * 
 * @author diewald
 */
public class ExplorationSolutionToDeployment extends TransformationStrategyBase<Deployment> {

	/**
	 * References the {@link ComponentArchitecture} that has been mapped onto a
	 * {@link PlatformArchitecture}.
	 */
	private ComponentArchitecture af3ComponentArchitecture;

	/** References the {@link PlatformArchitecture} for which this solution was generated. */
	private PlatformArchitecture af3PlatformArchitecture;

	/**
	 * Helper method to set the architecture model references that were provided as an input to the
	 * DSE.
	 */
	private <S extends EObject> void setArchModels(ExplorationTransformationInputs inputs)
			throws TransformationModuleException {
		// FIXME: Currently broken functionality, as the input set does not contain full
		// architecture models.
		af3ComponentArchitecture = null;// inputs.getEncoding(ComponentArchitecture.class);
		if(af3ComponentArchitecture == null) {
			throw new TransformationModuleException(this,
					"An ComponentArchitecture is required for this module to operate properly.");
		}

		// FIXME: Currently broken functionality, as the input set does not contain full
		// architecture models.
		af3PlatformArchitecture = null;// inputs.getEncoding(PlatformArchitecture.class);
		if(af3PlatformArchitecture == null) {
			throw new TransformationModuleException(this,
					"An PlatformArchitecture is required for this module to operate properly.");
		}
		// getPhysicalPlatformArchitecture(af3PlatformArchitecture);
	}

	/**
	 * {@inheritDoc}
	 * 
	 * @throws TransformationModuleException
	 *             see {@link ITransformationModule}.
	 */
	@Override
	public <N extends EObject, S extends EObject> Deployment transform(TransformationState state,
			TransformationContext context, ExplorationTransformationInputs inputs,
			SubClassToInstanceMap<N> transformedModels, EObject outputModel)
			throws TransformationModuleException {
		if(state == TransformationState.INITIALIZING) {
			@SuppressWarnings("unchecked") IMappingEncoding<ITaskAdapter<?>, ?, IMappingEntry<ITaskAdapter<?>, ?>> mapping =
					null;// inputs.getEncoding(IMappingEncoding.class);
			if(mapping == null) {
				throw new TransformationModuleException(this,
						"An IMappingEncoding is required for this module to operate properly.");
			}

			if(!isAdditionalArchRequiredByReplication(mapping) &&
					!isAdditionalArchRequiredByInstantiation(mapping)) {
				addBlacklistedModule(ComponentArchitectureTransformer.class);
			}

			setArchModels(inputs);
		}

		return super.transform(state, context, inputs, transformedModels, outputModel);
	}

	/**
	 * Predicate if replicas were generated and hence a modified {@link ComponentArchitecture} must
	 * be generated that is shipped with the generated {@link Deployment}.
	 */
	// TODO: Use "Component" as Generic.
	private boolean isAdditionalArchRequiredByReplication(
			IMappingEncoding<ITaskAdapter<?>, ?, IMappingEntry<ITaskAdapter<?>, ?>> mapping) {
		for(ITaskAdapter<?> deployedComponent : mapping.getRequesters()) {
			if(mapping.getAllocatedResourcesFor(deployedComponent).size() > 1) {
				return true;
			}
		}
		return false;
	}

	/**
	 * Predicate if "abstract" {@link Component}s were generated and hence a modified
	 * {@link ComponentArchitecture} must be generated that is shipped with the generated
	 * {@link Deployment}.
	 */
	// TODO: Use "Component" as Generic.
	private boolean isAdditionalArchRequiredByInstantiation(
			IMappingEncoding<ITaskAdapter<?>, ?, IMappingEntry<ITaskAdapter<?>, ?>> mapping) {
		for(ITaskAdapter<?> deployedComponent : mapping.getRequesters()) {
			if(deployedComponent.isInstantiation()) {
				return true;
			}
		}
		return false;
	}

	/** {@inheritDoc} */
	@Override
	public Class<Deployment> getOutputType() {
		return Deployment.class;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<Class<?>> getRequiredTypes() {
		Collection<Class<?>> inputTypes = new HashSet<>();
		inputTypes.add(ComponentArchitecture.class);
		inputTypes.add(PlatformArchitecture.class);
		inputTypes.add(InstantiatedTaskMappingEncoding.class);
		return inputTypes;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<Class<?>> getOptionalTypes() {
		return Collections.emptySet();
	}

	/** {@inheritDoc} */
	@Override
	public Collection<Class<? extends ITransformationModule<?>>> getParentModules() {
		return Collections.emptySet();
	}

	/** {@inheritDoc} */
	@Override
	public boolean isOptional() {
		return false;
	}

	/** {@inheritDoc} */
	@Override
	public TransformationState getActivationStage() {
		return TransformationState.INITIALIZING;
	}
}
